/**********************************************************
 **
 ** cmd_asm.c : manipulators for cmd_asm_t
 **
 ** Copyright (c) 2005, Kyle A. York
 ** All rights reserved
 **
 ***********************************************************/
#include <stdio.h>
#include <assert.h>
#include "../libpic12/pic_inst.h"
#include "cmdd.h"
#include "cmd_asm.h"

static void cmd_asm_dump_value(value_t val, const char *str, FILE *dst)
{
  if (val) {
    if (value_name_get(val)) {
      fprintf(dst, "%s%s", str, value_name_get(val));
    } else {
      fprintf(dst, "%s%u", str, (unsigned) value_const_get(val));
    }
  }
}

static void cmd_asm_dump(cmd_t cmd, FILE *dst)
{
  pic_optype_t type;

  fprintf(dst, "%s", pic_opcode_str(cmd_asm_op_get(cmd)));
  type = pic_optype_get(cmd_asm_op_get(cmd));
  switch (type) {
    case pic_optype_none:
      break;
    case pic_optype_f:
    case pic_optype_f_d:
    case pic_optype_f_b:
      cmd_asm_dump_value(cmd_asm_val_get(cmd), " ", dst);
      if (pic_optype_f_d == type) {
        fprintf(dst, ", %s", pic_opdst_str(cmd_asm_opdst_get(cmd)));
      } else if (pic_optype_f_b == type) {
        cmd_asm_dump_value(cmd_asm_n_get(cmd), ", ", dst);
      }
      break;
    case pic_optype_n:
    case pic_optype_tris:
      cmd_asm_dump_value(cmd_asm_n_get(cmd), " ", dst);
      break;
    case pic_optype_k:
      if (cmd_asm_lbl_get(cmd)) {
        fprintf(dst, " %s", label_name_get(cmd_asm_lbl_get(cmd)));
      } else {
        cmd_asm_dump_value(cmd_asm_n_get(cmd), " ", dst);
      }
      break;
    case pic_optype_macro:
      assert(0);
      break;
  }
}

static void cmd_asm_free(cmd_t cmd)
{
  cmd_asm_val_set(cmd, VALUE_NONE);
  cmd_asm_n_set(cmd, VALUE_NONE);
  cmd_asm_lbl_set(cmd, LABEL_NONE);
}

static const cmd_vtbl_t cmd_asm_vtbl = {
  cmd_asm_free,
  cmd_asm_dump
};

cmd_t cmd_asm_alloc(pic_opcode_t op, value_t val, 
    pic_opdst_t opdst, value_t n, label_t lbl, flag_t flags)
{
  cmd_t cmd;

  cmd = cmd_alloc(cmd_type_asm, &cmd_asm_vtbl);
  if (cmd) {
    struct cmd_ *ptr;

    ptr = cmd_element_seek(cmd, boolean_true);
    if (ptr) {
      ptr->u.op_asm.op    = op;
      ptr->u.op_asm.val   = VALUE_NONE;
      ptr->u.op_asm.opdst = opdst;
      ptr->u.op_asm.n     = VALUE_NONE;
      ptr->u.op_asm.lbl   = LABEL_NONE;
      ptr->u.op_asm.flags = flags;
      
      cmd_asm_val_set(cmd, val);
      cmd_asm_n_set(cmd, n);
      cmd_asm_lbl_set(cmd, lbl);
    }
  }
  return cmd;
}

pic_opcode_t cmd_asm_op_get(cmd_t cmd)
{
  struct cmd_ *ptr;

  ptr = cmd_element_seek(cmd, boolean_false);
  return (ptr) ? ptr->u.op_asm.op : pic_opcode_none;
}

void cmd_asm_op_set(cmd_t cmd, pic_opcode_t op)
{
  struct cmd_ *ptr;

  ptr = cmd_element_seek(cmd, boolean_true);
  if (ptr) {
    ptr->u.op_asm.op = op;
  }
}

value_t      cmd_asm_val_get(cmd_t cmd)
{
  struct cmd_ *ptr;

  ptr = cmd_element_seek(cmd, boolean_false);
  return (ptr) ? ptr->u.op_asm.val : VALUE_NONE;
}

void         cmd_asm_val_set(cmd_t cmd, value_t val)
{
  struct cmd_ *ptr;

  ptr = cmd_element_seek(cmd, boolean_true);
  if (ptr) {
    /* if a varaible is used in an assembly statement, mark it both
     * assigned *and* used */
    value_lock(val);
    value_assign_ct_bump(ptr->u.op_asm.val, ctr_bump_decr);
    value_use_ct_bump(ptr->u.op_asm.val, ctr_bump_decr);
    value_release(ptr->u.op_asm.val);
    ptr->u.op_asm.val = val;
    value_assign_ct_bump(val, ctr_bump_incr);
    value_use_ct_bump(val, ctr_bump_incr);
  }
}

pic_opdst_t  cmd_asm_opdst_get(cmd_t cmd)
{
  struct cmd_ *ptr;

  ptr = cmd_element_seek(cmd, boolean_false);
  return (ptr) ? ptr->u.op_asm.opdst : pic_opdst_none;
}

void cmd_asm_opdst_set(cmd_t cmd, pic_opdst_t dst)
{
  struct cmd_ *ptr;

  ptr = cmd_element_seek(cmd, boolean_true);
  if (ptr) {
    ptr->u.op_asm.opdst = dst;
  }
}

value_t      cmd_asm_n_get(cmd_t cmd)
{
  struct cmd_ *ptr;

  ptr = cmd_element_seek(cmd, boolean_false);
  return (ptr) ? ptr->u.op_asm.n : VALUE_NONE;
}

void         cmd_asm_n_set(cmd_t cmd, value_t n)
{
  struct cmd_ *ptr;

  ptr = cmd_element_seek(cmd, boolean_true);
  if (ptr) {
    value_lock(n);
    value_use_ct_bump(ptr->u.op_asm.n, ctr_bump_decr);
    value_release(ptr->u.op_asm.n);
    ptr->u.op_asm.n = n;
    value_use_ct_bump(n, ctr_bump_incr);
  }
}


label_t cmd_asm_lbl_get(cmd_t cmd)
{
  struct cmd_ *ptr;

  ptr = cmd_element_seek(cmd, boolean_false);
  return (ptr) ? ptr->u.op_asm.lbl : LABEL_NONE;
}

void cmd_asm_lbl_set(cmd_t cmd, label_t lbl)
{
  struct cmd_ *ptr;

  ptr = cmd_element_seek(cmd, boolean_true);
  if (ptr) {
    label_lock(lbl);
    label_usage_bump(ptr->u.op_asm.lbl, ctr_bump_decr);
    label_release(ptr->u.op_asm.lbl);
    ptr->u.op_asm.lbl = lbl;
    label_usage_bump(lbl, ctr_bump_incr);
  }
}

boolean_t cmd_asm_flag_test(cmd_t cmd, flag_t flag)
{
  struct cmd_ *ptr;

  ptr = cmd_element_seek(cmd, boolean_false);
  return (ptr) ? (ptr->u.op_asm.flags & flag) == flag : boolean_false;
}

flag_t cmd_asm_flag_get_all(cmd_t cmd)
{
  struct cmd_ *ptr;

  ptr = cmd_element_seek(cmd, boolean_false);
  return (ptr) ? ptr->u.op_asm.flags : 0;
}


